import platform
import os
from pathlib import Path
from typing import Union

import requests

from econuy.utils import get_project_root


def _get_binary(file_path: Union[str, os.PathLike, None] = None) -> None:
    if file_path is None:
        file_path = Path(get_project_root(), "utils")

    urls = {
        "Windows": "https://github.com/rxavier/econuy-extras/raw/main/econuy_extras/x13/windows/x13as.exe",
        "Darwin": "https://github.com/rxavier/econuy-extras/raw/main/econuy_extras/x13/darwin/x13as",
        "Linux": "https://github.com/rxavier/econuy-extras/raw/main/econuy_extras/x13/linux/x13as",
    }

    system_string = platform.system()
    if system_string not in urls.keys():
        raise ValueError("X13 binaries are only available for Windows, Darwin (macOS) or Linux.")
    if system_string == "Windows":
        suffix = ".exe"
    else:
        suffix = ""

    r = requests.get(urls[system_string])
    binary_path = Path(file_path, f"x13as{suffix}")
    with open(binary_path, "wb") as f:
        f.write(r.content)
    print("Download complete.")
    return binary_path


def _search_binary(
    start_path: Union[str, os.PathLike],
    n: int = 0,
    download_path: Union[str, os.PathLike, None] = None,
):
    """Recursively search for a the X13 binary starting from the n-parent folder of
    a supplied path."""
    if platform.system() == "Windows":
        search_term = "x13as.exe"
    else:
        search_term = "x13as"
    i = 0
    while i < n:
        i += 1
        start_path = os.path.dirname(start_path)
    try:
        final_path = [x for x in Path(start_path).rglob(search_term)][0].absolute().as_posix()
    except IndexError:
        print("X13 binary not found. Downloading appropiate binary for your system...")
        final_path = _get_binary(file_path=download_path).absolute().as_posix()

    return final_path



def original():
    from econuy.retrieval import (
        prices,
        fiscal_accounts,
        economic_activity,
        labor,
        external_sector,
        financial_sector,
        income,
    )

    return {
        "natacc_ind_con_nsa": {
            "description": "Cuentas nacionales: Oferta, precios constantes (mar-16, T)",
            "function": economic_activity.natacc_ind_con_nsa,
        },
        "natacc_gas_con_nsa": {
            "description": "Cuentas nacionales: Demanda, precios constantes (mar-16, T)",
            "function": economic_activity.natacc_gas_con_nsa,
        },
        "natacc_ind_cur_nsa": {
            "description": "Cuentas nacionales: Oferta, precios corrientes (mar-16, T)",
            "function": economic_activity.natacc_ind_cur_nsa,
        },
        "natacc_gas_cur_nsa": {
            "description": "Cuentas nacionales: Demanda, precios corrientes (mar-16, T)",
            "function": economic_activity.natacc_gas_cur_nsa,
        },
        "gdp_con_idx_sa": {
            "description": "PBI: índice real, desestacionalizado (mar-16, T)",
            "function": economic_activity.gdp_con_idx_sa,
        },
        "industrial_production": {
            "description": "Producción industrial: por división, agrupaciones y clases (ene-02, M)",
            "function": economic_activity.industrial_production,
        },
        "electricity": {
            "description": "Ventas de energía eléctrica por sector (ene-00, M)",
            "function": economic_activity.electricity,
        },
        "gasoline": {
            "description": "Ventas de nafta por departamento (ene-04, M)",
            "function": economic_activity.gasoline,
        },
        "diesel": {
            "description": "Ventas de gasoil por departamento (ene-04, M)",
            "function": economic_activity.diesel,
        },
        "cattle": {
            "description": "Faena de bovinos (2-ene-05, S)",
            "function": economic_activity.cattle,
        },
        "milk": {
            "description": "Remisión de leche a planta (ene-02, M)",
            "function": economic_activity.milk,
        },
        "cement": {
            "description": "Ventas de cemento (ene-90, M)",
            "function": economic_activity.cement,
        },
        "cpi": {
            "description": "Índice de precios al consumidor - IPC (ene-37, M)",
            "function": prices.cpi,
        },
        "nxr_monthly": {
            "description": "Tipo de cambio, interbancario, mensual (abr-72, M)",
            "function": prices.nxr_monthly,
        },
        "nxr_daily": {
            "description": "Tipo de cambio, interbancario, diario (3-ene-00, D)",
            "function": prices.nxr_daily,
        },
        "balance_gps": {
            "description": "Resultado fiscal: Sector público consolidado (ene-99, M)",
            "function": fiscal_accounts.balance_gps,
        },
        "balance_nfps": {
            "description": "Resultado fiscal: Sector público no financiero (ene-99, M)",
            "function": fiscal_accounts.balance_nfps,
        },
        "balance_cg-bps": {
            "description": "Resultado fiscal: Gobierno central-BPS (ene-99, M)",
            "function": fiscal_accounts.balance_cg_bps,
        },
        "balance_pe": {
            "description": "Resultado fiscal de las empresas públicas",
            "function": fiscal_accounts.balance_pe,
        },
        "balance_ancap": {
            "description": "Resultado fiscal: ANCAP (ene-01, M)",
            "function": fiscal_accounts.balance_ancap,
        },
        "balance_ute": {
            "description": "Resultado fiscal: UTE (ene-99, M)",
            "function": fiscal_accounts.balance_ute,
        },
        "balance_antel": {
            "description": "Resultado fiscal: ANTEL (ene-99, M)",
            "function": fiscal_accounts.balance_antel,
        },
        "balance_ose": {
            "description": "Resultado fiscal: OSE (ene-99, M)",
            "function": fiscal_accounts.balance_ose,
        },
        "tax_revenue": {
            "description": "Recaudación por impuesto (ene-82, M)",
            "function": fiscal_accounts.tax_revenue,
        },
        "public_debt_gps": {
            "description": "Deuda del sector público global: por plazo contractual, residual, moneda y residencia (dic-99, T)",
            "function": fiscal_accounts.public_debt_gps,
        },
        "public_debt_nfps": {
            "description": "Deuda del sector público no monetario: por plazo contractual, residual, moneda y residencia (dic-99, T)",
            "function": fiscal_accounts.public_debt_nfps,
        },
        "public_debt_cb": {
            "description": "Deuda del Banco Central: por plazo contractual, residual, moneda y residencia (dic-99, T)",
            "function": fiscal_accounts.public_debt_cb,
        },
        "public_assets": {
            "description": "Activos del sectro público (dic-99, T)",
            "function": fiscal_accounts.public_assets,
        },
        "labor_rates": {
            "description": "Tasas de actividad, empleo y desempleo por sexo (ene-06, M)",
            "function": labor.labor_rates,
        },
        "hours_worked": {
            "description": "Horas promedio trabajadas por sector (ene-06, M)",
            "function": labor.hours,
        },
        "nominal_wages": {
            "description": "Salarios nominales: total, públicos y privados (ene-68, M)",
            "function": labor.nominal_wages,
        },
        "real_wages": {
            "description": "Salarios reales: total, públicos y privados (ene-68, M)",
            "function": labor.real_wages,
        },
        "trade_x_prod_val": {
            "description": "Exportaciones por sector, valor (ene-00, M)",
            "function": external_sector.trade_x_prod_val,
        },
        "trade_x_prod_vol": {
            "description": "Exportaciones por sector, volumen (ene-05, M)",
            "function": external_sector.trade_x_prod_vol,
        },
        "trade_x_prod_pri": {
            "description": "Exportaciones por sector, precio (ene-05, M)",
            "function": external_sector.trade_x_prod_pri,
        },
        "trade_x_dest_val": {
            "description": "Exportaciones por destino, valor (ene-00, M)",
            "function": external_sector.trade_x_dest_val,
        },
        "trade_x_dest_vol": {
            "description": "Exportaciones por destino, volumen (ene-05, M)",
            "function": external_sector.trade_x_dest_vol,
        },
        "trade_x_dest_pri": {
            "description": "Exportaciones por destino, precio (ene-05, M)",
            "function": external_sector.trade_x_dest_pri,
        },
        "trade_m_sect_val": {
            "description": "Importaciones por tipo, valor (ene-00, M)",
            "function": external_sector.trade_m_sect_val,
        },
        "trade_m_sect_vol": {
            "description": "Importaciones por tipo, volumen (ene-05, M)",
            "function": external_sector.trade_m_sect_vol,
        },
        "trade_m_sect_pri": {
            "description": "Importaciones por tipo, precio (ene-05, M)",
            "function": external_sector.trade_m_sect_pri,
        },
        "trade_m_orig_val": {
            "description": "Importaciones por origen, valor (ene-00, M)",
            "function": external_sector.trade_m_orig_val,
        },
        "trade_m_orig_vol": {
            "description": "Importaciones por origen, volumen (ene-05, M)",
            "function": external_sector.trade_m_orig_vol,
        },
        "trade_m_orig_pri": {
            "description": "Importaciones por origen, precio (ene-05, M)",
            "function": external_sector.trade_m_orig_pri,
        },
        "rxr_official": {
            "description": "Tipos de cambio reales, BCU (ene-00, M)",
            "function": external_sector.rxr_official,
        },
        "reserves": {
            "description": "Reservas internacionales (26-jun-02, D)",
            "function": external_sector.reserves,
        },
        "reserves_changes": {
            "description": "Variación en las reservas del BCU por fuente (2-ene-13, D)",
            "function": external_sector.reserves_changes,
        },
        "deposits": {
            "description": "Depósitos en el sistema bancario (dic-98, M)",
            "function": financial_sector.deposits,
        },
        "credit": {
            "description": "Créditos del sistema bancario al sector no financiero (dic-98, M)",
            "function": financial_sector.credit,
        },
        "interest_rates": {
            "description": "Tasas de interés activas y pasivas (ene-98, M)",
            "function": financial_sector.interest_rates,
        },
        "call": {
            "description": "Tasa call a 1 día (02-ene-02, D)",
            "function": financial_sector.call_rate,
        },
        "sovereign_risk": {
            "description": "Uruguay Bond Index - riesgo soberano (01-ene-99, D)",
            "function": financial_sector.sovereign_risk,
        },
        "income_household": {
            "description": "Ingreso medio de los hogares sin valor ni aguinaldo (ene-06, M)",
            "function": income.income_household,
        },
        "income_capita": {
            "description": "Ingreso medio per cápita sin valor locativo ni aguinaldo (ene-06, M)",
            "function": income.income_capita,
        },
        "consumer_confidence": {
            "description": "Índice de Confianza del Consumidor (ago-07, M)",
            "function": income.consumer_confidence,
        },
    }


def custom():
    from econuy.retrieval import (
        prices,
        fiscal_accounts,
        economic_activity,
        labor,
        external_sector,
        financial_sector,
        international,
        regional,
    )

    return {
        "core_industrial": {
            "description": "Producción industrial: total, sin refinería y núcleo (ene-02, M [e])",
            "function": economic_activity.core_industrial,
        },
        "natacc_ind_con_nsa_long": {
            "description": "Cuentas nacionales: Oferta, precios constantes, serie empalmada (mar-88, T [e])",
            "function": economic_activity.natacc_ind_con_nsa_long,
        },
        "natacc_gas_con_nsa_long": {
            "description": "Cuentas nacionales: Demanda, precios constantes, serie empalmada (mar-88, T [e])",
            "function": economic_activity.natacc_gas_con_nsa_long,
        },
        "gdp_con_idx_sa_long": {
            "description": "PBI: índice real, desestacionalizado, serie empalmada (mar-88, T [e])",
            "function": economic_activity.gdp_con_idx_sa_long,
        },
        "gdp_con_nsa_long": {
            "description": "PBI: constante, serie empalmada (wp BCU 12-15) (mar-88, T [e])",
            "function": economic_activity.gdp_con_nsa_long,
        },
        "gdp_cur_nsa_long": {
            "description": "PBI: corriente, serie empalmada (wp BCU 12-15) (mar-97, T [e])",
            "function": economic_activity.gdp_cur_nsa_long,
        },
        "cpi_measures": {
            "description": "IPC transable, no transable, subyacente y residual (mar-97, M [e])",
            "function": prices.cpi_measures,
        },
        "balance_summary": {
            "description": "Resultado fiscal: Todas las agregaciones, inc. aj. FSS (ene-99, M [e])",
            "function": fiscal_accounts.balance_summary,
        },
        "net_public_debt": {
            "description": "Deuda neta del sector público global excl. encajes (dic-99, T [e])",
            "function": fiscal_accounts.net_public_debt,
        },
        "labor_rates_people": {
            "description": "Actividad, empleo y desempleo: series extendidas de tasas y personas (ene-91, M [e])",
            "function": labor.rates_people,
        },
        "trade_balance": {
            "description": "Balanza comercial por país (ene-00, M [e])",
            "function": external_sector.trade_balance,
        },
        "terms_of_trade": {
            "description": "Términos de intercambio (ene-05, M [e])",
            "function": external_sector.terms_of_trade,
        },
        "commodity_prices": {
            "description": "Precios de commodities seleccionados (ene-02, M [e])",
            "function": external_sector.commodity_prices,
        },
        "commodity_index": {
            "description": "Índice econuy de precios de commodities (ene-02, M [e])",
            "function": external_sector.commodity_index,
        },
        "rxr_custom": {
            "description": "Tipos de cambio reales, cálculos econuy (dic-79, M [e])",
            "function": external_sector.rxr_custom,
        },
        "bonds": {
            "description": "Rendimiento de bonos soberanos (02-jun-03, D [e])",
            "function": financial_sector.bonds,
        },
        "global_gdp": {
            "description": "PBI real desestacionalizado de EE.UU., Unión Europea, Japón y China (mar-47, T)",
            "function": international.gdp,
        },
        "global_stocks": {
            "description": "Índices bursátiles de EE.UU., Unión Europea, Japón y China (30-dic-27, D)",
            "function": international.stocks,
        },
        "global_policy_rates": {
            "description": "Tasas de interés de política monetaria de EE.UU., Eurozona, Japón y China (01-ene-46, D)",
            "function": international.policy_rates,
        },
        "global_long_rates": {
            "description": "Tasas de interés de largo plazo (10 años) de EE.UU., Alemania, Francia, España, Italia, Reino Unido, Japón y China (02-ene-62, D)",
            "function": international.long_rates,
        },
        "global_nxr": {
            "description": "Cotización de monedas de EE.UU., Eurozona, Japón y China (04-ene-71, D)",
            "function": international.nxr,
        },
        "regional_gdp": {
            "description": "PBI real desestacionalizado de Argentina y Brasil (mar-93, T)",
            "function": regional.gdp,
        },
        "regional_monthly_gdp": {
            "description": "PBI real mensual desestacionalizado de Argentina y Brasil (ene-03, M)",
            "function": regional.monthly_gdp,
        },
        "regional_cpi": {
            "description": "Índice de precios al consumo de Argentina y Brasil (ene-70, M)",
            "function": regional.cpi,
        },
        "regional_embi_spreads": {
            "description": "EMBI spread (JP Morgan) de Argentina, Brasil y Global (11-dic-98, D)",
            "function": regional.embi_spreads,
        },
        "regional_embi_yields": {
            "description": "EMBI yield (JP Morgan) de Argentina, Brasil y Global (11-dic-98, D)",
            "function": regional.embi_yields,
        },
        "regional_nxr": {
            "description": "Tipo de cambio nominal diario de Argentina y Brasil (9-abr-02, D)",
            "function": regional.nxr,
        },
        "regional_policy_rates": {
            "description": "Tasas de interés de política monetaria de Argentina y Brasil (4-jun-86, D)",
            "function": regional.policy_rates,
        },
        "regional_stocks": {
            "description": "Índices bursátiles en USD de Argentina y Brasil (9-abr-02, D)",
            "function": regional.stocks,
        },
        "regional_rxr": {
            "description": "Tipo de cambio real vs. EE.UU. (dic-70, M)",
            "function": regional.rxr,
        },
        "_lin_gdp": {
            "description": "PBI corriente acumulado mensual linealizado en pesos y USD con proyecciones",
            "function": economic_activity._lin_gdp,
        },
    }



import datetime as dt
import re
import tempfile
from pathlib import Path
from io import BytesIO
from os import listdir, path
from typing import List, Optional
from urllib.error import HTTPError, URLError

import pandas as pd
import numpy as np
import patoolib
import requests
from bs4 import BeautifulSoup
from opnieuw import retry
from pandas.tseries.offsets import MonthEnd

from econuy import transform
from econuy.core import Pipeline
from econuy.utils import metadata, get_project_root
from econuy.utils.sources import urls


def _natacc_retriever(
    url: str,
    nrows: int,
    skiprows: int,
    inf_adj: str,
    unit: str,
    seas_adj: str,
    colnames: List[str],
) -> pd.DataFrame:
    """Helper function. See any of the `natacc_...()` functions."""

    raw = (
        pd.read_excel(url, skiprows=skiprows, nrows=nrows, usecols="B:AAA", index_col=0)
        .dropna(how="all", axis=1)
        .dropna(how="all", axis=0)
        .T
    )
    raw.index = raw.index.str.replace("*", "", regex=True)
    raw.index = raw.index.str.replace(r"\bI \b", "3-", regex=True)
    raw.index = raw.index.str.replace(r"\bII \b", "6-", regex=True)
    raw.index = raw.index.str.replace(r"\bIII \b", "9-", regex=True)
    raw.index = raw.index.str.replace(r"\bIV \b", "12-", regex=True)
    raw.index = raw.index.str.replace(r"([0-9]+)(I)", r"\g<1>1", regex=True)
    raw.index = pd.to_datetime(raw.index, format="%m-%Y") + MonthEnd(1)
    raw.columns = colnames
    output = raw.apply(pd.to_numeric, errors="coerce")
    output.rename_axis(None, inplace=True)

    for col in output.columns:
        if "Importaciones" in col and all(output[col] < 0):
            output[col] = output[col] * -1

    metadata._set(
        output,
        area="Actividad económica",
        currency="UYU",
        inf_adj=inf_adj,
        unit=unit,
        seas_adj=seas_adj,
        ts_type="Flujo",
        cumperiods=1,
    )

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def natacc_ind_con_nsa() -> pd.DataFrame:
    """Get supply-side national accounts data in NSA constant prices, 2005-.

    Returns
    -------
    National accounts, supply side, constant prices, NSA : pd.DataFrame

    """
    colnames = [
        "Agropecuario, pesca y minería",
        "Industrias manufactureras",
        "Energía eléctrica, gas y agua",
        "Construcción",
        "Comercio, alojamiento y suministro de comidas y bebidas",
        "Transporte y almacenamiento, información y comunicaciones",
        "Servicios financieros",
        "Actividades profesionales y arrendamiento",
        "Actividades de administración pública",
        "Salud, educación, act. inmobiliarias y otros servicios",
        "Valor agregado a precios básicos",
        "Impuestos menos subvenciones",
        "Producto bruto interno",
    ]
    return _natacc_retriever(
        url=urls["natacc_ind_con_nsa"]["dl"]["main"],
        nrows=13,
        skiprows=7,
        inf_adj="Const. 2016",
        unit="Millones",
        seas_adj="NSA",
        colnames=colnames,
    )


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def natacc_gas_con_nsa() -> pd.DataFrame:
    """Get demand-side national accounts data in NSA constant prices, 2005-.

    Returns
    -------
    National accounts, demand side, constant prices, NSA : pd.DataFrame

    """
    colnames = [
        "Gasto de consumo: total",
        "Gasto de consumo: hogares",
        "Gasto de consumo: gobierno y ISFLH",
        "Formación bruta de capital",
        "Formación bruta de capital: fijo",
        "Variación de existencias",
        "Exportaciones de bienes y servicios",
        "Importaciones de bienes y servicios",
        "Producto bruto interno",
    ]
    return _natacc_retriever(
        url=urls["natacc_gas_con_nsa"]["dl"]["main"],
        nrows=9,
        skiprows=7,
        inf_adj="Const. 2016",
        unit="Millones",
        seas_adj="NSA",
        colnames=colnames,
    )


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def natacc_gas_cur_nsa() -> pd.DataFrame:
    """Get demand-side national accounts data in NSA current prices.

    Returns
    -------
    National accounts, demand side, current prices, NSA : pd.DataFrame

    """
    colnames = [
        "Gasto de consumo: total",
        "Gasto de consumo: hogares",
        "Gasto de consumo: gobierno y ISFLH",
        "Formación bruta de capital",
        "Formación bruta de capital: fijo",
        "Variación de existencias",
        "Exportaciones de bienes y servicios",
        "Importaciones de bienes y servicios",
        "Producto bruto interno",
    ]
    return _natacc_retriever(
        url=urls["natacc_gas_cur_nsa"]["dl"]["main"],
        nrows=9,
        skiprows=7,
        inf_adj="No",
        unit="Millones",
        seas_adj="NSA",
        colnames=colnames,
    )


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def natacc_ind_cur_nsa() -> pd.DataFrame:
    """Get supply-side national accounts data in NSA current prices, 2005-.

    Returns
    -------
    National accounts, supply side, current prices, NSA : pd.DataFrame

    """
    colnames = [
        "Agropecuario, pesca y minería",
        "Industrias manufactureras",
        "Energía eléctrica, gas y agua",
        "Construcción",
        "Comercio, alojamiento y suministro de comidas y bebidas",
        "Transporte y almacenamiento, información y comunicaciones",
        "Servicios financieros",
        "Actividades profesionales y arrendamiento",
        "Actividades de administración pública",
        "Salud, educación, act. inmobiliarias y otros servicios",
        "Valor agregado a precios básicos",
        "Impuestos menos subvenciones",
        "Producto bruto interno",
    ]
    return _natacc_retriever(
        url=urls["natacc_ind_cur_nsa"]["dl"]["main"],
        nrows=13,
        skiprows=7,
        inf_adj="No",
        unit="Millones",
        seas_adj="NSA",
        colnames=colnames,
    )


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def gdp_con_idx_sa() -> pd.DataFrame:
    """Get supply-side national accounts data in SA real index, 1997-.

    Returns
    -------
    National accounts, supply side, real index, SA : pd.DataFrame

    """
    colnames = ["Producto bruto interno"]
    return _natacc_retriever(
        url=urls["gdp_con_idx_sa"]["dl"]["main"],
        nrows=1,
        skiprows=7,
        inf_adj="Const. 2016",
        unit="2016=100",
        seas_adj="SA",
        colnames=colnames,
    )


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def natacc_ind_con_nsa_long(pipeline: Pipeline = None) -> pd.DataFrame:
    """Get supply-side national accounts data in NSA constant prices, 1988-.

    Three datasets with different base years, 1983, 2005 and 2016, are spliced
    in order to get to the result DataFrame.

    Returns
    -------
    National accounts, supply side, constant prices, NSA : pd.DataFrame

    """
    if pipeline is None:
        pipeline = Pipeline()
    pipeline.get("natacc_ind_con_nsa")
    data_16 = pipeline.dataset.copy()
    data_16.columns = data_16.columns.get_level_values(0)
    data_16["Otros servicios"] = (
        data_16["Servicios financieros"]
        + data_16["Actividades profesionales y arrendamiento"]
        + data_16["Actividades de administración pública"]
        + data_16["Salud, educación, act. inmobiliarias y otros servicios"]
    )
    data_16.drop(
        [
            "Valor agregado a precios básicos",
            "Servicios financieros",
            "Actividades profesionales y arrendamiento",
            "Actividades de administración pública",
            "Salud, educación, act. inmobiliarias y otros servicios",
            "Valor agregado a precios básicos",
        ],
        axis=1,
        inplace=True,
    )

    colnames = [
        "Actividades primarias",
        "Act. prim.: Agricultura, ganadería, caza y silvicultura",
        "Industrias manufactureras",
        "Suministro de electricidad, gas y agua",
        "Construcción",
        "Comercio, reparaciones, restaurantes y hoteles",
        "Transporte, almacenamiento y comunicaciones",
        "Otras actividades",
        "Otras actividades: SIFMI",
        "Impuestos menos subvenciones",
        "Producto bruto interno",
    ]
    data_05 = _natacc_retriever(
        url=urls["natacc_ind_con_nsa_long"]["dl"]["2005"],
        nrows=12,
        skiprows=9,
        inf_adj="Const. 2005",
        unit="Millones",
        seas_adj="NSA",
        colnames=colnames,
    )
    data_05.columns = data_05.columns.get_level_values(0)

    # In the following lines I rename and reorder the Other activities
    # column, and explicitly drop SIFMI because I have no way of adding them.
    # In order for this to be correct, I would have to substract the SIFMI
    # part from each of the other sectors, which I cannot do. Implicitly I'm
    # assuming that the share of SIFMI in each sector over time doesn't change,
    # which shouldn't affect growth rates.
    data_05["Otros servicios"] = data_05["Otras actividades"]
    data_05.drop(
        [
            "Act. prim.: Agricultura, ganadería, caza y silvicultura",
            "Otras actividades",
            "Otras actividades: SIFMI",
        ],
        axis=1,
        inplace=True,
    )
    data_05.columns = data_16.columns
    aux_index = list(dict.fromkeys(list(data_05.index) + list(data_16.index)))
    aux = data_16.reindex(aux_index)
    for quarter in reversed(aux_index):
        if aux.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            aux.loc[quarter, :] = (
                aux.loc[next_quarter, :] * data_05.loc[quarter, :] / data_05.loc[next_quarter, :]
            )
    aux = aux[
        [
            "Agropecuario, pesca y minería",
            "Industrias manufactureras",
            "Energía eléctrica, gas y agua",
            "Construcción",
            "Comercio, alojamiento y suministro de comidas y bebidas",
            "Transporte y almacenamiento, información y comunicaciones",
            "Otros servicios",
            "Impuestos menos subvenciones",
            "Producto bruto interno",
        ]
    ]

    data_83 = pd.read_excel(
        urls["natacc_ind_con_nsa_long"]["dl"]["1983"],
        skiprows=10,
        nrows=8,
        usecols="B:AAA",
        index_col=0,
    ).T
    data_83.index = pd.date_range(start="1988-03-31", freq="Q-DEC", periods=len(data_83))
    data_83["Impuestos menos subvenciones"] = np.nan
    data_83 = data_83[
        [
            "AGROPECUARIA",
            "INDUSTRIAS MANUFACTURERAS",
            "SUMINISTRO DE ELECTRICIDAD, GAS Y AGUA",
            "CONSTRUCCION",
            "COMERCIO, REPARACIONES, RESTAURANTES Y HOTELES",
            "TRANSPORTE, ALMACENAMIENTO Y COMUNICACIONES",
            "OTRAS (1)",
            "Impuestos menos subvenciones",
            "PRODUCTO INTERNO BRUTO",
        ]
    ]
    data_83.columns = aux.columns

    complete_index = list(dict.fromkeys(list(data_83.index) + list(aux.index)))
    output = aux.reindex(complete_index)
    for quarter in reversed(complete_index):
        if output.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            output.loc[quarter, :] = (
                output.loc[next_quarter, :]
                * data_83.loc[quarter, :]
                / data_83.loc[next_quarter, :]
            )

    metadata._set(
        output,
        area="Actividad económica",
        currency="UYU",
        inf_adj="Const. 2016",
        unit="Millones",
        seas_adj="NSA",
        ts_type="Flujo",
        cumperiods=1,
    )

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def natacc_gas_con_nsa_long(pipeline: Pipeline = None) -> pd.DataFrame:
    """Get demand-side national accounts data in NSA constant prices, 1988-.

    Three datasets with different base years, 1983, 2005 and 2016, are spliced
    in order to get to the result DataFrame.

    Returns
    -------
    National accounts, demand side, constant prices, NSA : pd.DataFrame

    """
    if pipeline is None:
        pipeline = Pipeline()
    pipeline.get("natacc_gas_con_nsa")
    data_16 = pipeline.dataset.copy()
    data_16.columns = data_16.columns.get_level_values(0)
    data_16.drop(["Variación de existencias"], axis=1, inplace=True)

    colnames = [
        "Gasto de consumo final",
        "Gasto de consumo final de hogares",
        "Gasto de consumo final del gobierno general",
        "Formación bruta de capital",
        "Formación bruta de capital fijo",
        "Sector público",
        "Sector privado",
        "Exportaciones",
        "Importaciones",
        "Producto bruto interno",
    ]
    data_05 = _natacc_retriever(
        url=urls["natacc_gas_con_nsa_long"]["dl"]["2005"],
        nrows=10,
        skiprows=9,
        inf_adj="Const. 2005",
        unit="Millones",
        seas_adj="NSA",
        colnames=colnames,
    )
    data_05.columns = data_05.columns.get_level_values(0)
    data_05.drop(["Sector público", "Sector privado"], axis=1, inplace=True)
    data_05.columns = data_16.columns
    aux_index = list(dict.fromkeys(list(data_05.index) + list(data_16.index)))
    aux = data_16.reindex(aux_index)
    for quarter in reversed(aux_index):
        if aux.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            aux.loc[quarter, :] = (
                aux.loc[next_quarter, :] * data_05.loc[quarter, :] / data_05.loc[next_quarter, :]
            )

    data_83 = pd.read_excel(
        urls["natacc_gas_con_nsa_long"]["dl"]["1983"],
        skiprows=10,
        nrows=11,
        usecols="B:AAA",
        index_col=0,
    ).T
    data_83.index = pd.date_range(start="1988-03-31", freq="Q-DEC", periods=len(data_83))
    data_83.drop(
        ["Sector público", "Sector privado", "Variación de existencias"], axis=1, inplace=True
    )
    data_83.columns = aux.columns

    complete_index = list(dict.fromkeys(list(data_83.index) + list(aux.index)))
    output = aux.reindex(complete_index)
    for quarter in reversed(complete_index):
        if output.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            output.loc[quarter, :] = (
                output.loc[next_quarter, :]
                * data_83.loc[quarter, :]
                / data_83.loc[next_quarter, :]
            )

    metadata._set(
        output,
        area="Actividad económica",
        currency="UYU",
        inf_adj="Const. 2016",
        unit="Millones",
        seas_adj="NSA",
        ts_type="Flujo",
        cumperiods=1,
    )

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def gdp_con_idx_sa_long(pipeline: Pipeline = None) -> pd.DataFrame:
    """Get demand-side national accounts data in NSA constant prices, 1988-.

    Three datasets with different base years, 1983, 2005 and 2016, are spliced
    in order to get to the result DataFrame.

    Returns
    -------
    National accounts, demand side, constant prices, NSA : pd.DataFrame

    """
    if pipeline is None:
        pipeline = Pipeline()
    pipeline.get("gdp_con_idx_sa")
    data_16 = pipeline.dataset.copy()
    data_16.columns = data_16.columns.get_level_values(0)

    colnames = [
        "Actividades primarias",
        "Act. prim.: Agricultura, ganadería, caza y silvicultura",
        "Industrias manufactureras",
        "Suministro de electricidad, gas y agua",
        "Construcción",
        "Comercio, reparaciones, restaurantes y hoteles",
        "Transporte, almacenamiento y comunicaciones",
        "Otras actividades",
        "Otras actividades: SIFMI",
        "Impuestos menos subvenciones",
        "Producto bruto interno",
    ]
    data_05 = _natacc_retriever(
        url=urls["gdp_con_idx_sa_long"]["dl"]["2005"],
        nrows=12,
        skiprows=9,
        inf_adj="Const. 2005",
        unit="Millones",
        seas_adj="SA",
        colnames=colnames,
    )
    data_05.columns = data_05.columns.get_level_values(0)
    data_05 = data_05[["Producto bruto interno"]]
    aux_index = list(dict.fromkeys(list(data_05.index) + list(data_16.index)))
    aux = data_16.reindex(aux_index)
    for quarter in reversed(aux_index):
        if aux.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            aux.loc[quarter, :] = (
                aux.loc[next_quarter, :] * data_05.loc[quarter, :] / data_05.loc[next_quarter, :]
            )

    data_83 = pd.read_excel(
        urls["gdp_con_idx_sa_long"]["dl"]["1983"],
        skiprows=10,
        nrows=8,
        usecols="B:AAA",
        index_col=0,
    ).T
    data_83.index = pd.date_range(start="1988-03-31", freq="Q-DEC", periods=len(data_83))
    data_83 = data_83[["PRODUCTO INTERNO BRUTO"]]
    data_83.columns = aux.columns

    complete_index = list(dict.fromkeys(list(data_83.index) + list(aux.index)))
    output = aux.reindex(complete_index)
    for quarter in reversed(complete_index):
        if output.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            output.loc[quarter, :] = (
                output.loc[next_quarter, :]
                * data_83.loc[quarter, :]
                / data_83.loc[next_quarter, :]
            )

    metadata._set(
        output,
        area="Actividad económica",
        currency="UYU",
        inf_adj="Const. 2016",
        unit="Millones",
        seas_adj="SA",
        ts_type="Flujo",
        cumperiods=1,
    )

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def gdp_con_nsa_long(pipeline: Pipeline = None) -> pd.DataFrame:
    """Get GDP data in NSA constant prices, 1988-.

    Four datasets with three different base years, 1983, 2005 and 2016, are
    spliced in order to get to the result DataFrame. It uses the BCU's working
    paper for retropolated GDP in current and constant prices for 2012-2015.

    Returns
    -------
    GDP, constant prices, NSA : pd.DataFrame

    """
    if pipeline is None:
        pipeline = Pipeline()
    pipeline.get("natacc_ind_con_nsa")
    data_16 = pipeline.dataset.copy()
    data_16.columns = data_16.columns.get_level_values(0)
    data_16 = data_16[["Producto bruto interno"]]

    data_12 = pd.read_csv(urls["gdp_con_nsa_long"]["dl"]["2012"], index_col=0, parse_dates=True)
    data_12 = data_12[["real"]]
    data_12.columns = data_16.columns

    aux = pd.concat([data_12, data_16], axis=0)

    colnames = ["Producto bruto interno"]
    data_05 = _natacc_retriever(
        url=urls["gdp_con_nsa_long"]["dl"]["2005"],
        nrows=2,
        skiprows=9,
        inf_adj="Const. 2005",
        unit="Millones",
        seas_adj="NSA",
        colnames=colnames,
    )
    data_05.columns = data_05.columns.get_level_values(0)
    aux_index = list(dict.fromkeys(list(data_05.index) + list(aux.index)))
    reaux = aux.reindex(aux_index)
    for quarter in reversed(aux_index):
        if reaux.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            reaux.loc[quarter, :] = (
                reaux.loc[next_quarter, :] * data_05.loc[quarter, :] / data_05.loc[next_quarter, :]
            )

    data_83 = pd.read_excel(
        urls["gdp_con_nsa_long"]["dl"]["1983"], skiprows=10, nrows=8, usecols="B:AAA", index_col=0
    ).T
    data_83.index = pd.date_range(start="1988-03-31", freq="Q-DEC", periods=len(data_83))
    data_83 = data_83[["PRODUCTO INTERNO BRUTO"]]
    data_83.columns = ["Producto bruto interno"]

    reaux_index = list(dict.fromkeys(list(data_83.index) + list(reaux.index)))
    output = reaux.reindex(reaux_index)
    for quarter in reversed(reaux_index):
        if output.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            output.loc[quarter, :] = (
                output.loc[next_quarter, :]
                * data_83.loc[quarter, :]
                / data_83.loc[next_quarter, :]
            )

    metadata._set(
        output,
        area="Actividad económica",
        currency="UYU",
        inf_adj="Const. 2016",
        unit="Millones",
        seas_adj="NSA",
        ts_type="Flujo",
        cumperiods=1,
    )

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def gdp_cur_nsa_long(pipeline: Pipeline = None) -> pd.DataFrame:
    """Get GDP data in NSA current prices, 1997-.

    Three datasets with two different base years, 2005 and 2016, are
    spliced in order to get to the result DataFrame. It uses the BCU's working
    paper for retropolated GDP in current and constant prices for 2012-2015.

    Returns
    -------
    GDP, current prices, NSA : pd.DataFrame

    """
    if pipeline is None:
        pipeline = Pipeline()
    pipeline.get("natacc_ind_cur_nsa")
    data_16 = pipeline.dataset.copy()
    data_16.columns = data_16.columns.get_level_values(0)
    data_16 = data_16[["Producto bruto interno"]]

    data_12 = pd.read_csv(urls["gdp_cur_nsa_long"]["dl"]["2012"], index_col=0, parse_dates=True)
    data_12 = data_12[["current"]]
    data_12.columns = data_16.columns

    aux = pd.concat([data_12, data_16], axis=0)

    colnames = ["Producto bruto interno"]
    data_05 = _natacc_retriever(
        url=urls["gdp_cur_nsa_long"]["dl"]["2005"],
        nrows=2,
        skiprows=9,
        inf_adj="No",
        unit="Millones",
        seas_adj="NSA",
        colnames=colnames,
    )
    data_05.columns = data_05.columns.get_level_values(0)
    aux_index = list(dict.fromkeys(list(data_05.index) + list(aux.index)))
    output = aux.reindex(aux_index)
    for quarter in reversed(aux_index):
        if output.loc[quarter, :].isna().all():
            next_quarter = quarter + MonthEnd(3)
            output.loc[quarter, :] = (
                output.loc[next_quarter, :]
                * data_05.loc[quarter, :]
                / data_05.loc[next_quarter, :]
            )

    metadata._set(
        output,
        area="Actividad económica",
        currency="UYU",
        inf_adj="No",
        unit="Millones",
        seas_adj="NSA",
        ts_type="Flujo",
        cumperiods=1,
    )

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def _lin_gdp(pipeline: Optional[Pipeline] = None):
    """Get nominal GDP data in UYU and USD with forecasts.

    Update nominal GDP data for use in the `transform.convert_gdp()` function.
    Get IMF forecasts for year of last available data point and the next
    year (for example, if the last period available at the BCU website is
    september 2019, fetch forecasts for 2019 and 2020).

    Parameters
    ----------
    pipeline : econuy.core.Pipeline or None, default None
        An instance of the econuy Pipeline class.

    Returns
    -------
    output : Pandas dataframe
        Quarterly GDP in UYU and USD with 1 year forecasts.

    """
    if pipeline is None:
        pipeline = Pipeline()

    pipeline.get(name="gdp_cur_nsa_long")
    data_uyu = pipeline.dataset
    # TODO: use Pipeline methods for these
    data_uyu = transform.rolling(data_uyu, window=4, operation="sum")
    data_usd = transform.convert_usd(data_uyu, pipeline=pipeline)

    data = [data_uyu, data_usd]
    last_year = data_uyu.index.max().year
    if data_uyu.index.max().month == 12:
        last_year += 1

    results = []
    for table, gdp in zip(["NGDP", "NGDPD"], data):
        table_url = (
            f"https://www.imf.org/en/Publications/WEO/weo-database/"
            f"2020/October/weo-report?c=298,&s={table},&sy="
            f"{last_year - 1}&ey={last_year + 1}&ssm=0&scsm=1&scc=0&"
            f"ssd=1&ssc=0&sic=0&sort=country&ds=.&br=1"
        )
        imf_data = pd.to_numeric(pd.read_html(table_url)[0].iloc[0, [5, 6, 7]])
        imf_data = imf_data.reset_index(drop=True)
        fcast = (
            gdp.loc[[dt.datetime(last_year - 1, 12, 31)]]
            .multiply(imf_data.iloc[1])
            .divide(imf_data.iloc[0])
        )
        fcast = fcast.rename(
            index={dt.datetime(last_year - 1, 12, 31): dt.datetime(last_year, 12, 31)}
        )
        next_fcast = (
            gdp.loc[[dt.datetime(last_year - 1, 12, 31)]]
            .multiply(imf_data.iloc[2])
            .divide(imf_data.iloc[0])
        )
        next_fcast = next_fcast.rename(
            index={dt.datetime(last_year - 1, 12, 31): dt.datetime(last_year + 1, 12, 31)}
        )
        fcast = fcast.append(next_fcast)
        gdp = gdp.append(fcast)
        results.append(gdp)

    output = pd.concat(results, axis=1)
    output = output.resample("Q-DEC").interpolate("linear").dropna(how="all")
    output.rename_axis(None, inplace=True)

    metadata._modify_multiindex(output, levels=[0], new_arrays=[["PBI UYU", "PBI USD"]])

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def industrial_production() -> pd.DataFrame:
    """Get industrial production data.

    Returns
    -------
    Monthly industrial production index : pd.DataFrame

    """
    name = "industrial_production"

    try:
        raw = pd.read_excel(urls[name]["dl"]["main"], skiprows=4, usecols="B:EM")
        weights = pd.read_excel(urls[name]["dl"]["weights"], skiprows=3, usecols="B:E").dropna(
            how="all"
        )
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            certificate = Path(get_project_root(), "utils", "files", "ine_certs.pem")
            r = requests.get(urls[name]["dl"]["main"], verify=certificate)
            raw = pd.read_excel(BytesIO(r.content), skiprows=4, usecols="B:EM")
            r = requests.get(urls[name]["dl"]["weights"], verify=certificate)
            weights = pd.read_excel(BytesIO(r.content), skiprows=3, usecols="B:E").dropna(
                how="all"
            )
        else:
            raise err
    proc = raw.dropna(how="any", subset=["Mes"]).dropna(thresh=100, axis=1)
    output = proc[~proc["Mes"].str.contains("PROM|Prom", regex=True)].drop("Mes", axis=1)
    output.index = pd.date_range(start="2002-01-31", freq="M", periods=len(output))

    column_names = []
    for c in output.columns[2:]:
        match = weights.loc[weights["division"] == c, "Denominación"]
        prefix = "Div_"
        if isinstance(match, pd.Series) and match.empty:
            prefix = "Agr_"
            match = weights.loc[weights["agrupacion"] == c, "Denominación"]
            if isinstance(match, pd.Series) and match.empty:
                prefix = "Cls_"
                match = weights.loc[weights["clase"] == c, "Denominación"]
        try:
            match = match.iloc[0]
        except AttributeError:
            pass
        match = (prefix + match.capitalize().strip())[:-1]
        match = re.sub(r"[\(\)]", "-", match)
        if len(match) > 60:
            match = match[:58] + "..."
        column_names.append(match)
    output.columns = [
        "Industrias manufactureras",
        "Industrias manufactureras sin refinería",
    ] + column_names
    output = output.apply(pd.to_numeric, errors="coerce")
    output.rename_axis(None, inplace=True)

    metadata._set(
        output,
        area="Actividad económica",
        currency="-",
        inf_adj="No",
        unit="2006=100",
        seas_adj="NSA",
        ts_type="Flujo",
        cumperiods=1,
    )

    return output


def core_industrial(pipeline: Optional[Pipeline] = None) -> pd.DataFrame:
    """
    Get total industrial production, industrial production excluding oil
    refinery and core industrial production.

    Parameters
    ----------
    pipeline : econuy.core.Pipeline or None, default None
        An instance of the econuy Pipeline class.

    Returns
    -------
    Measures of industrial production : pd.DataFrame

    """
    name = "core_industrial"

    if pipeline is None:
        pipeline = Pipeline()

    pipeline.get("industrial_production")
    data = pipeline.dataset

    try:
        weights = pd.read_excel(urls[name]["dl"]["weights"], skiprows=3).dropna(how="all")
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            certificate = Path(get_project_root(), "utils", "files", "ine_certs.pem")
            r = requests.get(urls[name]["dl"]["weights"], verify=certificate)
            weights = pd.read_excel(BytesIO(r.content), skiprows=3).dropna(how="all")
        else:
            raise err
    weights = weights.rename(
        columns={
            "Unnamed: 5": "Pond. división",
            "Unnamed: 6": "Pond. agrupación",
            "Unnamed: 7": "Pond. clase",
        }
    )
    other_foods = (
        weights.loc[weights["clase"] == 1549]["Pond. clase"].values[0]
        * weights.loc[(weights["agrupacion"] == 154) & (weights["clase"] == 0)][
            "Pond. agrupación"
        ].values[0]
        * weights.loc[(weights["division"] == 15) & (weights["agrupacion"] == 0)][
            "Pond. división"
        ].values[0]
        / 1000000
    )
    pulp = (
        weights.loc[weights["clase"] == 2101]["Pond. clase"].values[0]
        * weights.loc[(weights["division"] == 21) & (weights["agrupacion"] == 0)][
            "Pond. división"
        ].values[0]
        / 10000
    )
    output = data.loc[:, ["Industrias manufactureras", "Industrias manufactureras sin refinería"]]
    exclude = (
        data.loc[:, "Cls_Elaboración de productos alimenticios n.c.p"] * other_foods
        + data.loc[:, "Cls_Pulpa de madera, papel y cartón"] * pulp
    )
    core = data["Industrias manufactureras sin refinería"] - exclude
    core = pd.concat([core], keys=["Núcleo industrial"], names=["Indicador"], axis=1)
    output = pd.concat([output, core], axis=1)
    output = transform.rebase(output, start_date="2006-01-01", end_date="2006-12-31")
    output.rename_axis(None, inplace=True)

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def cattle() -> pd.DataFrame:
    """Get weekly cattle slaughter data.

    Returns
    -------
    Weekly cattle slaughter : pd.DataFrame

    """
    name = "cattle"
    try:
        output = pd.read_excel(urls[name]["dl"]["main"], skiprows=8, usecols="C:H")
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            certificate = Path(get_project_root(), "utils", "files", "inac_certs.pem")
            r = requests.get(urls[name]["dl"]["main"], verify=certificate)
            output = pd.read_excel(BytesIO(r.content), skiprows=8, usecols="C:H")
        else:
            raise err

    output.index = pd.date_range(start="2005-01-02", freq="W", periods=len(output))
    output.rename_axis(None, inplace=True)

    metadata._set(
        output,
        area="Actividad económica",
        currency="-",
        inf_adj="No",
        unit="Cabezas",
        seas_adj="NSA",
        ts_type="Flujo",
        cumperiods=1,
    )

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def milk() -> pd.DataFrame:
    """Get monthly milk production in farms data.

    Returns
    -------
    Monhtly milk production in farms : pd.DataFrame

    """
    name = "milk"

    r = requests.get(urls[name]["dl"]["main"])
    soup = BeautifulSoup(r.content, features="lxml")
    link = soup.find_all(href=re.compile(".xls"))[0]
    raw = pd.read_excel(link["href"], skiprows=11, skipfooter=4)
    output = raw.iloc[:, 2:].drop(0, axis=0)
    output = pd.melt(output, id_vars="Año/ Mes")[["value"]].dropna()
    output.index = pd.date_range(start="2002-01-31", freq="M", periods=len(output))
    output = output.apply(pd.to_numeric)
    output.columns = ["Remisión de leche a planta"]
    output.rename_axis(None, inplace=True)

    metadata._set(
        output,
        area="Actividad económica",
        currency="-",
        inf_adj="No",
        unit="Miles de litros",
        seas_adj="NSA",
        ts_type="Flujo",
        cumperiods=1,
    )

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def cement() -> pd.DataFrame:
    """Get monthly cement sales data.

    Returns
    -------
    Monthly cement sales : pd.DataFrame

    """
    name = "cement"

    output = pd.read_excel(
        urls[name]["dl"]["main"], skiprows=2, usecols="B:E", index_col=0, skipfooter=1
    )
    output.index = output.index + MonthEnd(0)
    output.columns = ["Exportaciones", "Mercado interno", "Total"]
    output.rename_axis(None, inplace=True)

    metadata._set(
        output,
        area="Actividad económica",
        currency="-",
        inf_adj="No",
        unit="Toneladas",
        seas_adj="NSA",
        ts_type="Flujo",
        cumperiods=1,
    )
    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def diesel() -> pd.DataFrame:
    """
    Get diesel sales by department data.

    This retrieval function requires the unrar binaries to be found in your
    system.

    Returns
    -------
    Monthly diesel dales : pd.DataFrame

    """
    name = "diesel"

    temp_rar = tempfile.NamedTemporaryFile(suffix=".rar").name
    with open(temp_rar, "wb") as f:
        r = requests.get(urls[name]["dl"]["main"])
        soup = BeautifulSoup(r.content, features="lxml")
        rar_url = soup.find_all(href=re.compile("gas%20oil"))[0]
        f.write(requests.get(rar_url["href"]).content)
    with tempfile.TemporaryDirectory() as temp_dir:
        patoolib.extract_archive(temp_rar, outdir=temp_dir, verbosity=-1)
        xls = [x for x in listdir(temp_dir) if x.endswith(".xls")][0]
        path_temp = path.join(temp_dir, xls)
        raw = pd.read_excel(
            path_temp, sheet_name="vta gas oil por depto", skiprows=2, usecols="C:W"
        )
        raw.index = pd.date_range(start="2004-01-31", freq="M", periods=len(raw))
        raw.columns = list(raw.columns.str.replace("\n", " "))[:-1] + ["Total"]
        output = raw
    output.rename_axis(None, inplace=True)

    metadata._set(
        output,
        area="Actividad económica",
        currency="-",
        inf_adj="No",
        unit="m3",
        seas_adj="NSA",
        ts_type="Flujo",
        cumperiods=1,
    )

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def gasoline() -> pd.DataFrame:
    """
    Get gasoline sales by department data.

    This retrieval function requires the unrar binaries to be found in your
    system.

    Returns
    -------
    Monthly gasoline dales : pd.DataFrame

    """
    name = "gasoline"

    temp_rar = tempfile.NamedTemporaryFile(suffix=".rar").name
    with open(temp_rar, "wb") as f:
        r = requests.get(urls[name]["dl"]["main"])
        soup = BeautifulSoup(r.content, features="lxml")
        rar_url = soup.find_all(href=re.compile("gasolina"))[0]
        f.write(requests.get(rar_url["href"]).content)
    with tempfile.TemporaryDirectory() as temp_dir:
        patoolib.extract_archive(temp_rar, outdir=temp_dir, verbosity=-1)
        xls = [x for x in listdir(temp_dir) if x.endswith(".xls")][0]
        path_temp = path.join(temp_dir, xls)
        raw = pd.read_excel(
            path_temp, sheet_name="vta gasolinas por depto", skiprows=2, usecols="C:W"
        )
        raw.index = pd.date_range(start="2004-01-31", freq="M", periods=len(raw))
        raw.columns = list(raw.columns.str.replace("\n", " "))[:-1] + ["Total"]
        output = raw
    output.rename_axis(None, inplace=True)

    metadata._set(
        output,
        area="Actividad económica",
        currency="-",
        inf_adj="No",
        unit="m3",
        seas_adj="NSA",
        ts_type="Flujo",
        cumperiods=1,
    )

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def electricity() -> pd.DataFrame:
    """
    Get electricity sales by sector data.

    This retrieval function requires the unrar binaries to be found in your
    system.

    Returns
    -------
    Monthly electricity dales : pd.DataFrame

    """
    name = "electricity"

    temp_rar = tempfile.NamedTemporaryFile(suffix=".rar").name
    with open(temp_rar, "wb") as f:
        r = requests.get(urls[name]["dl"]["main"])
        soup = BeautifulSoup(r.content, features="lxml")
        rar_url = soup.find_all(href=re.compile("Facturaci[%A-z0-9]+sector"))[0]
        f.write(requests.get(rar_url["href"]).content)
    with tempfile.TemporaryDirectory() as temp_dir:
        patoolib.extract_archive(temp_rar, outdir=temp_dir, verbosity=-1)
        xls = [x for x in listdir(temp_dir) if x.endswith(".xls")][0]
        path_temp = path.join(temp_dir, xls)
        raw = pd.read_excel(path_temp, sheet_name="fact ee", skiprows=2, usecols="C:J")
        raw.index = pd.date_range(start="2000-01-31", freq="M", periods=len(raw))
        raw.columns = raw.columns.str.capitalize()
        output = raw
    output.rename_axis(None, inplace=True)

    metadata._set(
        output,
        area="Actividad económica",
        currency="-",
        inf_adj="No",
        unit="MWh",
        seas_adj="NSA",
        ts_type="Flujo",
        cumperiods=1,
    )

    return output
